"""
Abstract interpreter and type definitions for the Nada DSL auditing
component.
"""

# pylint: disable=attribute-defined-outside-init
# pylint: disable=too-many-lines
# pylint: disable=too-few-public-methods
from __future__ import annotations
from typing import Union, Tuple, Sequence
import ast


class Metaclass(type):
    """
    Metaclass for the :obj:`Constant`, :obj:`Public`, and :obj:`Secret`
    classes, enabling comparison of derived classes and conversions
    between them.
    """

    def shape(cls: type, cls_: type = None) -> type:
        """
        Return the shape of a class that is an instance of this
        metaclass.

        >>> PublicInteger.shape().__name__
        'Public'
        >>> AbstractInteger.shape(Constant).__name__
        'Integer'
        """
        # pylint: disable=too-many-return-statements
        if cls_ is not None:
            if issubclass(cls, AbstractInteger):
                if cls_ == Constant:
                    return Integer
                if cls_ == Public:
                    return PublicInteger
                if cls_ == Secret:
                    return SecretInteger
            if issubclass(cls, AbstractBoolean):
                if cls_ == Constant:
                    return Boolean
                if cls_ == Public:
                    return PublicBoolean
                if cls_ == Secret:
                    return SecretBoolean

            return None

        if issubclass(cls, Constant):
            return Constant

        if issubclass(cls, Public):
            return Public

        if issubclass(cls, Secret):
            return Secret

        return None

    def __lt__(cls: type, other: type) -> bool:
        if issubclass(cls, Constant):
            return issubclass(other, (Public, Secret))

        if issubclass(cls, Public):
            return issubclass(other, Secret)

        return False

    def __le__(cls: type, other: type) -> bool:
        if issubclass(cls, Public):
            return not issubclass(other, Constant)

        if issubclass(cls, Secret):
            return issubclass(other, Secret)

        return True


class Constant(metaclass=Metaclass):
    """
    Class from which classes for constants are derived.

    >>> Constant <= Constant
    True
    >>> Constant < Secret
    True
    >>> Constant < Secret
    True
    >>> Secret <= Constant
    False
    >>> min(Constant, Public).__name__
    'Constant'
    """


class Public(metaclass=Metaclass):
    """
    Class from which classes for public values are derived.

    >>> Public < Secret
    True
    >>> Public <= Constant
    False
    >>> Public < Public
    False
    >>> min(Secret, Public).__name__
    'Public'
    """


class Secret(metaclass=Metaclass):
    """
    Class from which classes for constants are derived.

    >>> Constant < Secret
    True
    >>> Public < Secret
    True
    >>> Secret < Secret
    False
    >>> max([Secret, Public, Constant]).__name__
    'Secret'
    """


class Abstract(metaclass=Metaclass):
    """
    Base class for abstract interpreter values. All more specific abstract
    value types are derived from this class.

    The attributes of this class are also used as global aggregators of the
    signature components (parties, inputs, and outputs) during abstract execution.

    >>> Abstract.initialize()
    >>> party = Party("party")
    >>> input = Input("input", party)
    >>> output = Output(PublicInteger(input), "output", party)
    >>> [
    ...     list(map(lambda entry: type(entry).__name__, component))
    ...     for component in Abstract.signature()
    ... ]
    [['Party'], ['Input'], ['Output']]
    """

    # pylint: disable=missing-function-docstring
    parties = None
    inputs = None
    outputs = None
    context = None
    analysis = None

    @staticmethod
    def initialize(context=None):
        Abstract.parties = []
        Abstract.inputs = []
        Abstract.outputs = []
        Abstract.context = context if context is not None else {}
        Abstract.analysis = {"add": 0, "mul": 0, "cmp": 0, "eq": 0, "ne": 0, "ife": 0}

    @staticmethod
    def party(party: Party):
        Abstract.parties.append(party)

    @staticmethod
    def input(input: Input):  # pylint: disable=redefined-builtin
        Abstract.inputs.append(input)

    @staticmethod
    def output(output: Output):
        Abstract.outputs.append(output)

    @staticmethod
    def signature() -> Tuple[list[Party], list[Input], list[Output]]:
        return (Abstract.parties, Abstract.inputs, Abstract.outputs)

    def __init__(self: Abstract, cls: type = None):
        self.value = None
        if cls is not None:
            self.__class__ = cls


class Party(Abstract):
    """
    Abstract interpreter values corresponding to parties.

    >>> isinstance(Party("party"), Party)
    True

    If arguments having incorrect types are supplied to the constructor, an
    exception is raised.

    >>> Party(123)
    Traceback (most recent call last):
      ...
    TypeError: name parameter must be a string
    """

    def __init__(self: Party, name: str):
        super().__init__()

        type(self).party(self)

        if not isinstance(name, str):
            raise TypeError("name parameter must be a string")

        self.name = name


class Input(Abstract):
    """
    Abstract interpreter values corresponding to inputs.

    >>> isinstance(Input("input", Party("party")), Input)
    True

    If arguments having incorrect types are supplied to the constructor, an
    exception is raised.

    >>> Input(123, Party("party"))
    Traceback (most recent call last):
      ...
    TypeError: name parameter must be a string
    >>> Input("input", 456)
    Traceback (most recent call last):
      ...
    TypeError: party parameter must be a Party object
    """

    def __init__(self: Input, name: str, party: Party):
        super().__init__()

        type(self).input(self)

        if not isinstance(name, str):
            raise TypeError("name parameter must be a string")

        if not isinstance(party, Party):
            raise TypeError("party parameter must be a Party object")

        self.name = name
        self.party = party

    def _value(self) -> int:
        """
        Retrieve the value for this input when performing concrete interpretation.
        """
        return self.context.get(self.name, None)


class Output(Abstract):
    """
    Abstract interpreter values corresponding to outputs.

    >>> party = Party("party")
    >>> input = Input("input", party)
    >>> isinstance(Output(PublicInteger(input), "output", party), Output)
    True

    If arguments having incorrect types are supplied to the constructor, an
    exception is raised.

    >>> Output(123, "output", party)
    Traceback (most recent call last):
      ...
    TypeError: output value must be a PublicInteger or a SecretInteger
    >>> Output(PublicInteger(input), 123, party)
    Traceback (most recent call last):
      ...
    TypeError: name parameter must be a string
    >>> Output(PublicInteger(input), "output", 123)
    Traceback (most recent call last):
      ...
    TypeError: party parameter must be a Party object
    """

    def __init__(
        self: Output,
        value: Union[PublicInteger, SecretInteger],
        name: str,
        party: Party,
    ):
        super().__init__()

        type(self).output(self)

        if not isinstance(value, (PublicInteger, SecretInteger)):
            raise TypeError("output value must be a PublicInteger or a SecretInteger")

        if not isinstance(name, str):
            raise TypeError("name parameter must be a string")

        if not isinstance(party, Party):
            raise TypeError("party parameter must be a Party object")

        self.value = value
        self.name = name
        self.party = party

        # Store signature in this object (the first such instance constructed)
        # and reset the :obj:`Abstract` class attribute aggregators.
        self.final = (type(self).parties, type(self).inputs, type(self).outputs)
        self.analysis = Abstract.analysis


class AbstractInteger(Abstract):
    """
    Abstract interpreter values corresponding to all integers types.
    This class is only used by derived classes and is not exported.
    """

    def __init__(
        self: Output,
        input: Input = None,  # pylint: disable=redefined-builtin
        value: int = None,
    ):
        super().__init__()

        self.input = input
        self.value = self.input._value() if input is not None else value
        if input is not None:
            if not hasattr(input, "_type"):
                setattr(input, "_type", None)
            input._type = type(self)

    def __add__(
        self: AbstractInteger, other: Union[int, AbstractInteger]
    ) -> AbstractInteger:
        """
        Addition of abstract values that are instances of integer classes.
        The table below presents the output type for each combination
        of argument types.

        +-------------------+-------------------+-------------------+
        |     ``self``      |     ``other``     |    **result**     |
        +-------------------+-------------------+-------------------+
        |    ``Integer``    |    ``Integer``    |    ``Integer``    |
        +-------------------+-------------------+-------------------+
        |    ``Integer``    | ``PublicInteger`` | ``PublicInteger`` |
        +-------------------+-------------------+-------------------+
        |    ``Integer``    | ``SecretInteger`` | ``SecretInteger`` |
        +-------------------+-------------------+-------------------+
        | ``PublicInteger`` |    ``Integer``    | ``PublicInteger`` |
        +-------------------+-------------------+-------------------+
        | ``PublicInteger`` | ``PublicInteger`` | ``PublicInteger`` |
        +-------------------+-------------------+-------------------+
        | ``PublicInteger`` | ``SecretInteger`` | ``SecretInteger`` |
        +-------------------+-------------------+-------------------+
        | ``SecretInteger`` |    ``Integer``    | ``SecretInteger`` |
        +-------------------+-------------------+-------------------+
        | ``SecretInteger`` | ``PublicInteger`` | ``SecretInteger`` |
        +-------------------+-------------------+-------------------+
        | ``SecretInteger`` | ``SecretInteger`` | ``SecretInteger`` |
        +-------------------+-------------------+-------------------+

        >>> x = Input("x", Party("a"))
        >>> y = Input("y", Party("b"))
        >>> type(PublicInteger(x) + PublicInteger(y)).__name__
        'PublicInteger'
        >>> type(PublicInteger(x) + SecretInteger(y)).__name__
        'SecretInteger'
        >>> type(SecretInteger(x) + PublicInteger(y)).__name__
        'SecretInteger'
        >>> type(SecretInteger(x) + SecretInteger(y)).__name__
        'SecretInteger'

        The addition operator supports ``0`` as a base case in order to
        accommodate the built-in :obj:`sum` function.

        >>> type(sum([PublicInteger(x), SecretInteger(y)])).__name__
        'SecretInteger'
        >>> type(sum([SecretInteger(x), SecretInteger(y)])).__name__
        'SecretInteger'
        >>> type(sum([PublicInteger(x), PublicInteger(y)])).__name__
        'PublicInteger'

        If a value of an incompatible type is supplied to an overloaded
        operator, an exception is raised.

        >>> PublicInteger(x) + 123
        Traceback (most recent call last):
          ...
        TypeError: expecting Integer, PublicInteger, or SecretInteger
        """
        if (
            isinstance(other, int) and other == 0
        ):  # Base case for compatibility with :obj:`sum`.
            result = Abstract(type(self))
            result.value = self.value
            return result

        if not isinstance(other, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        result = Abstract(max(type(self), type(other)))

        Abstract.analysis["add"] += 1

        result.value = None
        if self.value is not None and other.value is not None:
            result.value = self.value + other.value

        return result

    def __radd__(
        self: AbstractInteger, other: Union[int, AbstractInteger]
    ) -> AbstractInteger:
        """
        Addition for cases in which the left-hand argument is not an instance
        of a class derived from this class.
        """
        return self.__add__(other)

    def __sub__(self: AbstractInteger, other: AbstractInteger) -> AbstractInteger:
        """
        Subtraction of abstract values that are instances of integer classes.
        """
        if not isinstance(other, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        result = Abstract(max(type(self), type(other)))

        result.value = None
        if self.value is not None and other.value is not None:
            result.value = self.value - other.value

        return result

    def __rsub__(self: AbstractInteger, other: AbstractInteger) -> AbstractInteger:
        """
        Subtraction for cases in which the left-hand argument is not an instance
        of a class derived from this class.
        """
        return self.__sub__(other)

    def __neg__(self: AbstractInteger) -> AbstractInteger:
        """
        Negation of abstract values that are instances of integer classes.

        >>> x = Input("x", Party("a"))
        >>> type(-Integer(x)).__name__
        'Integer'
        >>> type(-PublicInteger(x)).__name__
        'PublicInteger'
        >>> type(-SecretInteger(x)).__name__
        'SecretInteger'
        """
        result = Abstract(type(self))

        result.value = None
        if self.value is not None:
            result.value = -self.value

        return result

    def __mul__(self: AbstractInteger, other: AbstractInteger) -> AbstractInteger:
        """
        Multipliaction of abstract values that are instances of integer
        classes. The table below presents the output type for each
        combination of argument types.

        +-------------------+-------------------+-------------------+
        |     ``self``      |     ``other``     |    **result**     |
        +-------------------+-------------------+-------------------+
        |    ``Integer``    |    ``Integer``    |    ``Integer``    |
        +-------------------+-------------------+-------------------+
        |    ``Integer``    | ``PublicInteger`` | ``PublicInteger`` |
        +-------------------+-------------------+-------------------+
        |    ``Integer``    | ``SecretInteger`` | ``SecretInteger`` |
        +-------------------+-------------------+-------------------+
        | ``PublicInteger`` |    ``Integer``    | ``PublicInteger`` |
        +-------------------+-------------------+-------------------+
        | ``PublicInteger`` | ``PublicInteger`` | ``PublicInteger`` |
        +-------------------+-------------------+-------------------+
        | ``PublicInteger`` | ``SecretInteger`` | ``SecretInteger`` |
        +-------------------+-------------------+-------------------+
        | ``SecretInteger`` |    ``Integer``    | ``SecretInteger`` |
        +-------------------+-------------------+-------------------+
        | ``SecretInteger`` | ``PublicInteger`` | ``SecretInteger`` |
        +-------------------+-------------------+-------------------+
        | ``SecretInteger`` | ``SecretInteger`` | ``SecretInteger`` |
        +-------------------+-------------------+-------------------+

        >>> x = Input("x", Party("a"))
        >>> y = Input("y", Party("b"))
        >>> type(PublicInteger(x) + PublicInteger(y)).__name__
        'PublicInteger'
        >>> type(PublicInteger(x) + SecretInteger(y)).__name__
        'SecretInteger'
        >>> type(SecretInteger(x) + PublicInteger(y)).__name__
        'SecretInteger'
        >>> type(SecretInteger(x) + SecretInteger(y)).__name__
        'SecretInteger'

        The addition operator supports ``0`` as a base case in order to
        accommodate the built-in :obj:`sum` function.

        >>> type(sum([PublicInteger(x), SecretInteger(y)])).__name__
        'SecretInteger'
        >>> type(sum([SecretInteger(x), SecretInteger(y)])).__name__
        'SecretInteger'
        >>> type(sum([PublicInteger(x), PublicInteger(y)])).__name__
        'PublicInteger'

        If a value of an incompatible type is supplied to an overloaded
        operator, an exception is raised.

        >>> PublicInteger(x) + 123
        Traceback (most recent call last):
          ...
        TypeError: expecting Integer, PublicInteger, or SecretInteger
        """
        if not isinstance(other, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        result = Abstract(max(type(self), type(other)))

        Abstract.analysis["mul"] += 1

        result.value = None
        if self.value is not None and other.value is not None:
            result.value = self.value * other.value

        return result

    def __rmul__(self: AbstractInteger, other: AbstractInteger) -> AbstractInteger:
        """
        Addition for cases in which the left-hand argument is not an instance
        of a class derived from this class.
        """
        return self.__mul__(other)

    def __lt__(self: AbstractInteger, other: AbstractInteger) -> AbstractInteger:
        """
        Comparison of abstract values that are instances of integer
        classes. The table below presents the output type for each
        combination of argument types.

        +-------------------+-------------------+-------------------+
        |     ``self``      |     ``other``     |    **result**     |
        +-------------------+-------------------+-------------------+
        |    ``Integer``    |    ``Integer``    |    ``Boolean``    |
        +-------------------+-------------------+-------------------+
        |    ``Integer``    | ``PublicInteger`` | ``PublicBoolean`` |
        +-------------------+-------------------+-------------------+
        |    ``Integer``    | ``SecretInteger`` | ``SecretBoolean`` |
        +-------------------+-------------------+-------------------+
        | ``PublicInteger`` |    ``Integer``    | ``PublicBoolean`` |
        +-------------------+-------------------+-------------------+
        | ``PublicInteger`` | ``PublicInteger`` | ``PublicBoolean`` |
        +-------------------+-------------------+-------------------+
        | ``PublicInteger`` | ``SecretInteger`` | ``SecretBoolean`` |
        +-------------------+-------------------+-------------------+
        | ``SecretInteger`` |    ``Integer``    | ``SecretBoolean`` |
        +-------------------+-------------------+-------------------+
        | ``SecretInteger`` | ``PublicInteger`` | ``SecretBoolean`` |
        +-------------------+-------------------+-------------------+
        | ``SecretInteger`` | ``SecretInteger`` | ``SecretBoolean`` |
        +-------------------+-------------------+-------------------+

        >>> x = Input("x", Party("a"))
        >>> y = Input("y", Party("b"))
        >>> type(PublicInteger(x) < PublicInteger(y)).__name__
        'PublicBoolean'
        >>> type(PublicInteger(x) < SecretInteger(y)).__name__
        'SecretBoolean'
        >>> type(Integer(x) < Integer(y)).__name__
        'Boolean'
        >>> type(SecretInteger(x) < Integer(y)).__name__
        'SecretBoolean'
        >>> type(Integer(x) < PublicInteger(y)).__name__
        'PublicBoolean'

        If a value of an incompatible type is supplied to an overloaded
        operator, an exception is raised.

        >>> PublicInteger(x) < 123
        Traceback (most recent call last):
          ...
        TypeError: expecting Integer, PublicInteger, or SecretInteger
        """
        if not isinstance(other, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        shape = max(type(self), type(other)).shape()
        result = (AbstractBoolean.shape(shape))()
        Abstract.analysis["cmp"] += 1
        result.value = None
        if self.value is not None and other.value is not None:
            result.value = self.value < other.value

        return result

    def __le__(self: AbstractInteger, other: AbstractInteger) -> AbstractBoolean:
        """
        Comparison of abstract values that are instances of integer
        classes. See :obj:`AbstractInteger.__lt__` for details and examples.
        """
        if not isinstance(other, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        shape = max(type(self), type(other)).shape()
        result = (AbstractBoolean.shape(shape))()

        Abstract.analysis["cmp"] += 1

        result.value = None
        if self.value is not None and other.value is not None:
            result.value = self.value <= other.value

        return result

    def __gt__(self: AbstractInteger, other: AbstractInteger) -> AbstractBoolean:
        """
        Comparison of abstract values that are instances of integer
        classes. See :obj:`AbstractInteger.__lt__` for details and examples.
        """
        if not isinstance(other, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        shape = max(type(self), type(other)).shape()
        result = (AbstractBoolean.shape(shape))()

        Abstract.analysis["cmp"] += 1

        result.value = None
        if self.value is not None and other.value is not None:
            result.value = self.value > other.value

        return result

    def __ge__(self: AbstractInteger, other: AbstractInteger) -> AbstractBoolean:
        """
        Comparison of abstract values that are instances of integer
        classes. See :obj:`AbstractInteger.__lt__` for details and examples.
        """
        if not isinstance(other, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        shape = max(type(self), type(other)).shape()
        result = (AbstractBoolean.shape(shape))()

        Abstract.analysis["cmp"] += 1

        result.value = None
        if self.value is not None and other.value is not None:
            result.value = self.value >= other.value

        return result

    def __eq__(self: AbstractInteger, other: AbstractInteger) -> AbstractBoolean:
        """
        Comparison of abstract values that are instances of integer
        classes. See :obj:`AbstractInteger.__lt__` for details and examples.
        """
        if not isinstance(other, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        shape = max(type(self), type(other)).shape()
        result = (AbstractBoolean.shape(shape))()

        Abstract.analysis["eq"] += 1

        result.value = None
        if self.value is not None and other.value is not None:
            result.value = self.value == other.value

        return result

    def __ne__(self: AbstractInteger, other: AbstractInteger) -> AbstractBoolean:
        """
        Comparison of abstract values that are instances of integer
        classes. See :obj:`AbstractInteger.__lt__` for details and examples.
        """
        if not isinstance(other, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        shape = max(type(self), type(other)).shape()
        result = (AbstractBoolean.shape(shape))()

        Abstract.analysis["ne"] += 1

        result.value = None
        if self.value is not None and other.value is not None:
            result.value = self.value != other.value

        return result


class Integer(AbstractInteger, Constant):
    """
    Abstract values corresponding to constant integers.

    >>> Integer < PublicInteger
    True
    >>> Integer < SecretInteger
    True
    >>> Integer.shape().__name__
    'Constant'

    Instances of this class support the use of some built-in operators.

    >>> x = Input("x", Party("a"))
    >>> y = Input("y", Party("b"))
    >>> type(Integer(x) + Integer(y)).__name__
    'Integer'
    >>> type(Integer(x) * Integer(y)).__name__
    'Integer'
    >>> type(Integer(y) + PublicInteger(x)).__name__
    'PublicInteger'
    >>> type(SecretInteger(x) + Integer(y)).__name__
    'SecretInteger'
    >>> type(PublicInteger(x) * Integer(y)).__name__
    'PublicInteger'
    >>> type(Integer(y) * SecretInteger(x)).__name__
    'SecretInteger'

    The addition operator supports ``0`` as a base case in order to
    accommodate the built-in :obj:`sum` function.

    >>> type(sum([Integer(x), Integer(y)])).__name__
    'Integer'
    >>> type(sum([Integer(x), SecretInteger(y)])).__name__
    'SecretInteger'
    >>> type(sum([PublicInteger(x), Integer(y)])).__name__
    'PublicInteger'

    Concrete interpretation (with explicit values) is also supported if
    those values are present in the aggregate context being maintained
    using the static class attributes.

    >>> Abstract.context = {'x': 123, 'y': 456}
    >>> r = Integer(x, 123) + SecretInteger(y, 456)
    >>> r.value
    579
    >>> r = PublicInteger(x, 123) + Integer(y, 456)
    >>> r.value
    579
    >>> r = SecretInteger(x, 123) * Integer(y, 456)
    >>> r.value
    56088
    >>> r = Integer(x, 123) * PublicInteger(y, 456)
    >>> r.value
    56088

    If a value of an incompatible type is supplied to an overloaded
    operator, an exception is raised.

    >>> Integer(x) + 123
    Traceback (most recent call last):
      ...
    TypeError: expecting Integer, PublicInteger, or SecretInteger
    >>> Integer(x) * 123
    Traceback (most recent call last):
      ...
    TypeError: expecting Integer, PublicInteger, or SecretInteger
    >>> 123 * Integer(x)
    Traceback (most recent call last):
      ...
    TypeError: expecting Integer, PublicInteger, or SecretInteger
    """


class PublicInteger(AbstractInteger, Public):
    """
    Abstract values corresponding to public integers.

    >>> Integer < PublicInteger
    True
    >>> PublicInteger < SecretInteger
    True
    >>> PublicInteger.shape().__name__
    'Public'

    Instances of this class support the use of some built-in operators.

    >>> x = Input("x", Party("a"))
    >>> y = Input("y", Party("b"))
    >>> type(PublicInteger(x) + PublicInteger(y)).__name__
    'PublicInteger'
    >>> type(PublicInteger(x) + SecretInteger(y)).__name__
    'SecretInteger'
    >>> type(SecretInteger(x) + PublicInteger(y)).__name__
    'SecretInteger'
    >>> type(PublicInteger(x) * PublicInteger(y)).__name__
    'PublicInteger'
    >>> type(PublicInteger(x) * SecretInteger(y)).__name__
    'SecretInteger'
    >>> type(SecretInteger(x) * PublicInteger(y)).__name__
    'SecretInteger'

    The addition operator supports ``0`` as a base case in order to
    accommodate the built-in :obj:`sum` function.

    >>> type(sum([PublicInteger(x), SecretInteger(y)])).__name__
    'SecretInteger'
    >>> type(sum([PublicInteger(x), PublicInteger(y)])).__name__
    'PublicInteger'

    Concrete interpretation (with explicit values) is also supported if
    those values are present in the aggregate context being maintained
    using the static class attributes.

    >>> Abstract.context = {'x': 123, 'y': 456}
    >>> r = PublicInteger(x, 123) + SecretInteger(y, 456)
    >>> r.value
    579
    >>> r = PublicInteger(x, 123) + PublicInteger(y, 456)
    >>> r.value
    579
    >>> r = PublicInteger(x, 123) * SecretInteger(y, 456)
    >>> r.value
    56088
    >>> r = PublicInteger(x, 123) * PublicInteger(y, 456)
    >>> r.value
    56088

    If a value of an incompatible type is supplied to an overloaded
    operator, an exception is raised.

    >>> PublicInteger(x) + 123
    Traceback (most recent call last):
      ...
    TypeError: expecting Integer, PublicInteger, or SecretInteger
    >>> PublicInteger(x) * 123
    Traceback (most recent call last):
      ...
    TypeError: expecting Integer, PublicInteger, or SecretInteger
    >>> 123 * PublicInteger(x)
    Traceback (most recent call last):
      ...
    TypeError: expecting Integer, PublicInteger, or SecretInteger
    """


class SecretInteger(AbstractInteger, Secret):
    """
    Abstract interpreter values corresponding to secret integers.

    >>> SecretInteger < Integer
    False
    >>> PublicInteger < SecretInteger
    True
    >>> SecretInteger.shape().__name__
    'Secret'

    Instances of this class support the use of some built-in operators.

    >>> x = Input("x", Party("a"))
    >>> y = Input("y", Party("b"))
    >>> type(Integer(x) + SecretInteger(y)).__name__
    'SecretInteger'
    >>> type(SecretInteger(x) + PublicInteger(y)).__name__
    'SecretInteger'
    >>> type(SecretInteger(x) + SecretInteger(y)).__name__
    'SecretInteger'
    >>> type(PublicInteger(x) * SecretInteger(y)).__name__
    'SecretInteger'
    >>> type(SecretInteger(x) * Integer(y)).__name__
    'SecretInteger'
    >>> type(SecretInteger(x) * SecretInteger(y)).__name__
    'SecretInteger'

    Concrete interpretation (with explicit values) is also supported if
    those values are present in the aggregate context being maintained
    using the static class attributes.

    >>> Abstract.context = {'x': 123, 'y': 456}
    >>> r = SecretInteger(x, 123) * SecretInteger(y, 456)
    >>> r.value
    56088

    If a value of an incompatible type is supplied to an overloaded
    operator, an exception is raised.

    >>> SecretInteger(x) + 123
    Traceback (most recent call last):
      ...
    TypeError: expecting Integer, PublicInteger, or SecretInteger
    >>> SecretInteger(x) * 123
    Traceback (most recent call last):
      ...
    TypeError: expecting Integer, PublicInteger, or SecretInteger
    >>> 123 * SecretInteger(x)
    Traceback (most recent call last):
      ...
    TypeError: expecting Integer, PublicInteger, or SecretInteger
    """


class AbstractBoolean(Abstract):
    """
    Abstract interpreter values corresponding to all boolean types.
    This class is only used by derived classes and is not exported.
    """

    def __init__(
        self: Output,
        input: Input = None,  # pylint: disable=redefined-builtin
        value: int = None,
    ):
        super().__init__()

        self.input = input
        self.value = self.input._value() if input is not None else value
        if input is not None:
            if not hasattr(input, "_type"):
                setattr(input, "_type", None)
            input._type = type(self)

    def if_else(
        self: AbstractBoolean, true: AbstractInteger, false: AbstractInteger
    ) -> AbstractInteger:
        """
        Ternary (*i.e.*, conditional) operator. The table below presents
        the output type for each combination of argument types.

        +-------------------+------------------------+-------------------+
        |     ``self``      | ``true`` and ``false`` |    **result**     |
        +-------------------+------------------------+-------------------+
        |    ``Boolean``    |      ``Integer``       |    ``Integer``    |
        +-------------------+------------------------+-------------------+
        |    ``Boolean``    |   ``PublicInteger``    | ``PublicInteger`` |
        +-------------------+------------------------+-------------------+
        |    ``Boolean``    |   ``SecretInteger``    | ``SecretInteger`` |
        +-------------------+------------------------+-------------------+
        | ``PublicBoolean`` |      ``Integer``       | ``PublicInteger`` |
        +-------------------+------------------------+-------------------+
        | ``PublicBoolean`` |   ``PublicInteger``    | ``PublicInteger`` |
        +-------------------+------------------------+-------------------+
        | ``PublicBoolean`` |   ``SecretInteger``    | ``SecretInteger`` |
        +-------------------+------------------------+-------------------+
        | ``SecretBoolean`` |      ``Integer``       | ``SecretInteger`` |
        +-------------------+------------------------+-------------------+
        | ``SecretBoolean`` |   ``PublicInteger``    | ``SecretInteger`` |
        +-------------------+------------------------+-------------------+
        | ``SecretBoolean`` |   ``SecretInteger``    | ``SecretInteger`` |
        +-------------------+------------------------+-------------------+

        >>> x = Input("x", Party("a"))
        >>> y = Input("y", Party("b"))
        >>> z = Input("z", Party("c"))
        >>> type(PublicBoolean(x).if_else(PublicInteger(y), PublicInteger(z))).__name__
        'PublicInteger'

        If a value of an incompatible type is supplied to an overloaded
        operator, an exception is raised.

        >>> PublicInteger(x) + 123
        Traceback (most recent call last):
          ...
        TypeError: expecting Integer, PublicInteger, or SecretInteger
        """
        if not isinstance(true, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        if not isinstance(false, (Integer, PublicInteger, SecretInteger)):
            raise TypeError("expecting Integer, PublicInteger, or SecretInteger")

        shape = max([type(self), type(true), type(false)]).shape()
        result = (AbstractInteger.shape(shape))()

        Abstract.analysis["ife"] += 1

        result.value = None
        if (
            self.value is not None
            and true.value is not None
            and false.value is not None
        ):
            result.value = true.value if self.value else false.value

        return result


class Boolean(AbstractBoolean, Constant):
    """
    Abstract values corresponding to constant boolean values.
    """


class PublicBoolean(AbstractBoolean, Public):
    """
    Abstract values corresponding to public boolean values.
    """


class SecretBoolean(AbstractBoolean, Secret):
    """
    Abstract values corresponding to secret boolean values.
    """


def signature(source: str) -> Tuple[list[Party], list[Input], list[Output]]:
    """
    Return the signature of the supplied Nada program (represented as a
    string). The signature consists of three lists: the program's (1) parties,
    (2) inputs, and (3) outputs.

    >>> source = '\\n'.join([
    ...     'from nada_dsl import *',
    ...     '',
    ...     'def nada_main():',
    ...     '    party1 = Party(name="Party1")',
    ...     '    a = SecretInteger(Input(name="a", party=party1))',
    ...     '    b = SecretInteger(Input(name="b", party=party1))',
    ...     '',
    ...     '    c = a * b',
    ...     '',
    ...     '    return [Output(c, "c", party1), Output(c, "d", party1)]'
    ... ])
    >>> (ps, ins, outs) = signature(source)
    >>> [p.name for p in ps]
    ['Party1']
    >>> [i.name for i in ins]
    ['a', 'b']
    >>> [o.name for o in outs]
    ['c', 'd']

    If the supplied strign is not a valid Nada program, an exception is raised.

    >>> source = '\\n'.join([
    ...     'from nada_dsl import *',
    ...     'def nada_main():',
    ...     '    pass'
    ... ])
    >>> signature(source)
    Traceback (most recent call last):
      ...
    ValueError: nada_main must return a sequence of outputs
    >>> source = '\\n'.join([
    ...     'def nada_main():',
    ...     '    pass'
    ... ])
    >>> signature(source)
    Traceback (most recent call last):
      ...
    ValueError: first statement must be: from nada_dsl import *
    >>> source = '\\n'.join([
    ...     'from nada_dsl import *',
    ...     'def foo():',
    ...     '    pass'
    ... ])
    >>> signature(source)
    Traceback (most recent call last):
      ...
    ValueError: nada_main must be defined
    """
    root = ast.parse(source)

    if (
        len(root.body) == 0
        or not isinstance(root.body[0], ast.ImportFrom)
        or len(root.body[0].names) != 1
        or root.body[0].names[0].name != "*"
        or root.body[0].module != "nada_dsl"
    ):
        raise ValueError("first statement must be: from nada_dsl import *")

    # Adjust the import statement and add a statement that resets the static
    # class attributes being used for aggregation.
    root.body[0].module = "nada_dsl.audit"
    # root.body.append(ast.Expr(ast.Call(ast.Name('nada_main', ast.Load()), [], [])))
    root.body.append(
        ast.Expr(
            ast.Call(
                ast.Attribute(
                    ast.Name("Abstract", ast.Load()), "initialize", ast.Load()
                ),
                [],
                [],
            )
        )
    )
    ast.fix_missing_locations(root)

    # Execute the program (introducing the main function into the context).
    context = {}
    exec(compile(root, "", "exec"), context)  # pylint: disable=exec-used
    if "nada_main" not in context:
        raise ValueError("nada_main must be defined")

    # Perform abstract execution of the main function and return the signature of
    # the result.
    outputs = context["nada_main"]()
    if (
        isinstance(outputs, Sequence)
        and len(outputs) > 0
        and all(isinstance(output, Output) for output in outputs)
    ):
        return Abstract.signature()

    raise ValueError("nada_main must return a sequence of outputs")



"""
Common classes and functions for the Nada DSL auditing component
(used across different static analysis submodules).
"""

# pylint: disable=wildcard-import,invalid-name
# pylint: disable=too-few-public-methods
from __future__ import annotations
import ast


class Rule(Exception):
    """
    Base class for violations of rules defined by static analysis submodules.
    """


class RuleInAncestor:
    """
    Rule attribute value that indicates that the Nada DSL rule attribute of an
    :obj:`ast` node is determined by an ancestor node's Nada DSL rule attribute.
    """


class SyntaxRestriction(Rule):
    """
    Base class for violations of syntax restrictions defined by static analysis
    submodules.
    """


class TypeErrorRoot(TypeError):
    """
    Class for type errors that are not caused by other type errors.
    """


class TypeInParent:
    """
    Type attribute value that indicates that the Nada DSL type attribute of an
    :obj:`ast` node is determined by the parent node's Nada DSL type attribute.
    """


class Feedback:
    """
    Feedback aggregator (used throughout the recursive static analysis
    algorithms in order to collect all created exceptions).
    """

    def __init__(self: Feedback):
        self.exceptions = []

    def __call__(self: Feedback, exception: Exception):
        self.exceptions.append(exception)
        return exception


def typeerror_demote(t):
    """
    Demote a direct-cause type error to a generic (possibly indirect) type error.
    """
    if isinstance(t, TypeErrorRoot):
        return TypeError(str(t))

    return t


def audits(node, key, value=None, default=None, delete=False):
    """
    Set, update, or delete an :obj:`ast` node's static analysis attribute.
    """
    # pylint: disable=protected-access
    if not hasattr(node, "_audits"):
        setattr(node, "_audits", {})

    if value is None:
        value = node._audits.get(key, default)
        if delete and key in node._audits:
            del node._audits[key]
        return value

    node._audits[key] = value

    return None


def rules_no_restriction(a, recursive=False):
    """
    Delete the rule attributes of an :obj:`ast` node (and possibly those of its
    descendants).
    """
    if recursive:
        for a_ in ast.walk(a):
            audits(a_, "rules", delete=True)
    else:
        audits(a, "rules", delete=True)


def unify(t_a, t_b):
    """
    Perform basic type unification of Python types.
    """
    if t_a == t_b:
        return t_a

    if t_a.__name__ == "list" and t_b.__name__ == "list":
        if hasattr(t_a, "__args__") and len(t_a.__args__) == 1:
            if hasattr(t_b, "__args__"):
                if len(t_b.__args__) == 1:
                    return unify(t_a.__args__[0], t_b.__args__[0])
                return None
            return t_a

    return None



"""
Package exports.
"""

from nada_dsl.audit import *
from nada_dsl.source_ref import *
from nada_dsl.nada_types.scalar_types import *
from nada_dsl.nada_types.generics import *
from nada_dsl.nada_types.collections import *
from nada_dsl.nada_types.function import *
from nada_dsl.program_io import *
from nada_dsl.compiler_frontend import nada_compile



"""Nada Collection type definitions."""

from dataclasses import dataclass
from typing import Any, Dict, Generic, List
import typing

from sortedcontainers import SortedDict

from nada_dsl.ast_util import (
    AST_OPERATIONS,
    BinaryASTOperation,
    MapASTOperation,
    TupleAccessorASTOperation,
    NTupleAccessorASTOperation,
    NewASTOperation,
    ObjectAccessorASTOperation,
    ReduceASTOperation,
    UnaryASTOperation,
)
from nada_dsl.nada_types import DslType

# Wildcard import due to non-zero types
from nada_dsl.nada_types.scalar_types import *  # pylint: disable=W0614:wildcard-import
from nada_dsl.source_ref import SourceRef
from nada_dsl.errors import (
    IncompatibleTypesError,
    InvalidTypeError,
    NotAllowedException,
)
from nada_dsl.nada_types.function import NadaFunction, create_nada_fn
from nada_dsl.nada_types.generics import U, T, R
from . import AllTypes, OperationType


def is_primitive_integer(nada_type_str: str):
    """TODO: Autogenerate this method"""
    return (
        nada_type_str
        in (
            "Integer",
            "PublicInteger",
            "SecretInteger",
            "UnsignedInteger",
            "PublicUnsignedInteger",
            "SecretUnsignedInteger",
        ),
    )


@dataclass
class Map(Generic[T, R]):
    """The Map operation"""

    child: OperationType
    fn: NadaFunction[T, R]
    source_ref: SourceRef

    def __init__(
        self,
        child: OperationType,
        fn: NadaFunction[T, R],
        source_ref: SourceRef,
    ):
        self.id = OperationId.next()
        self.child = child
        self.fn = fn
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store MP in AST"""
        AST_OPERATIONS[self.id] = MapASTOperation(
            id=self.id,
            child=self.child.child.id,
            fn=self.fn.id,
            source_ref=self.source_ref,
            ty=ty,
        )


@dataclass
class Reduce(Generic[T, R]):
    """The Nada Reduce operation."""

    child: OperationType
    fn: NadaFunction[T, R]
    initial: R
    source_ref: SourceRef

    def __init__(
        self,
        child: OperationType,
        fn: NadaFunction[T, R],
        initial: R,
        source_ref: SourceRef,
    ):
        self.id = OperationId.next()
        self.child = child
        self.fn = fn
        self.initial = initial
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store a reduce object in AST"""
        AST_OPERATIONS[self.id] = ReduceASTOperation(
            id=self.id,
            child=self.child.child.id,
            fn=self.fn.id,
            initial=self.initial.child.id,
            source_ref=self.source_ref,
            ty=ty,
        )


class TupleType(NadaType):
    """Marker type for Tuples."""

    is_compound = True

    def __init__(self, left_type: NadaType, right_type: NadaType):
        self.left_type = left_type
        self.right_type = right_type

    def instantiate(self, child_or_value):
        return Tuple(child_or_value, self.left_type, self.right_type)

    def to_mir(self) -> proto_ty.NadaType:
        """Convert a tuple object into a Nada type."""
        return proto_ty.NadaType(
            tuple=proto_ty.Tuple(
                left=self.left_type.to_mir(), right=self.right_type.to_mir()
            )
        )


def _generate_accessor(ty: Any, accessor: Any) -> DslType:
    if hasattr(ty, "ty") and ty.ty.is_literal():  # TODO: fix
        raise TypeError("Literals are not supported in accessors")
    return ty.instantiate(accessor)


@dataclass
class Tuple(Generic[T, U], DslType):
    """The Tuple type"""

    left_type: NadaType
    right_type: NadaType

    def __init__(self, child, left_type: NadaType, right_type: NadaType):
        self.left_type = left_type
        self.right_type = right_type
        self.child = child
        super().__init__(self.child)

    @classmethod
    def new(cls, left_value: T, right_value: U) -> "Tuple[T, U]":
        """Constructs a new Tuple."""
        return Tuple(
            left_type=left_value.type(),
            right_type=right_value.type(),
            child=TupleNew(
                child=(left_value, right_value),
                source_ref=SourceRef.back_frame(),
            ),
        )

    @classmethod
    def generic_type(cls, left_type: U, right_type: T) -> TupleType:
        """Returns the generic type for this Tuple"""
        return TupleType(left_type=left_type, right_type=right_type)

    @property
    def left(self) -> DslType:
        """The left element of the  Tuple."""
        accessor = TupleAccessor(
            index=0,
            child=self,
            source_ref=SourceRef.back_frame(),
        )

        return _generate_accessor(self.left_type, accessor)

    @property
    def right(self) -> DslType:
        """The right element of the Tuple."""
        accessor = TupleAccessor(
            index=1,
            child=self,
            source_ref=SourceRef.back_frame(),
        )

        return _generate_accessor(self.right_type, accessor)

    def type(self):
        """Metatype for Tuple"""
        return TupleType(self.left_type, self.right_type)


@dataclass
class TupleAccessor:
    """Accessor for Tuple"""

    child: Tuple
    index: int
    source_ref: SourceRef

    def __init__(
        self,
        child: Tuple,
        index: int,
        source_ref: SourceRef,
    ):
        self.id = OperationId.next()
        self.child = child
        self.index = index
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store this accessor in the AST."""
        AST_OPERATIONS[self.id] = TupleAccessorASTOperation(
            id=self.id,
            source=self.child.child.id,
            index=self.index,
            source_ref=self.source_ref,
            ty=ty,
        )


class NTupleType(NadaType):
    """Marker type for NTuples."""

    is_compound = True

    def __init__(self, types: List[NadaType]):
        self.types = types

    def instantiate(self, child_or_value):
        return NTuple(child_or_value, self.types)

    def to_mir(self) -> proto_ty.NadaType:
        """Convert a tuple object into a Nada type."""
        return proto_ty.NadaType(
            ntuple=proto_ty.Ntuple(fields=[ty.to_mir() for ty in self.types])
        )


@dataclass
class NTuple(DslType):
    """The NTuple type"""

    types: List[Any]

    def __init__(self, child, types: List[Any]):
        self.types = types
        self.child = child
        super().__init__(self.child)

    @classmethod
    def new(cls, values: List[Any]) -> "NTuple":
        """Constructs a new NTuple."""
        types = [value.type() for value in values]
        return NTuple(
            types=types,
            child=NTupleNew(
                child=values,
                source_ref=SourceRef.back_frame(),
            ),
        )

    def __getitem__(self, index: int) -> DslType:
        if index >= len(self.types):
            raise IndexError(f"Invalid index {index} for NTuple.")

        accessor = NTupleAccessor(
            index=index,
            child=self,
            source_ref=SourceRef.back_frame(),
        )

        return _generate_accessor(self.types[index], accessor)

    def type(self):
        """Metatype for NTuple"""
        return NTupleType(self.types)


@dataclass
class NTupleAccessor:
    """Accessor for NTuple"""

    child: NTuple
    index: int
    source_ref: SourceRef

    def __init__(
        self,
        child: NTuple,
        index: int,
        source_ref: SourceRef,
    ):
        self.id = OperationId.next()
        self.child = child
        self.index = index
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store this accessor in the AST."""
        AST_OPERATIONS[self.id] = NTupleAccessorASTOperation(
            id=self.id,
            source=self.child.child.id,
            index=self.index,
            source_ref=self.source_ref,
            ty=ty,
        )


class ObjectType(NadaType):
    """Marker type for Objects."""

    is_compound = True

    def __init__(self, types: Dict[str, NadaType]):
        self.types = types

    def to_mir(self) -> proto_ty.NadaType:
        """Convert an object into a Nada type."""
        fields = SortedDict({name: ty.to_mir() for name, ty in self.types.items()})
        return proto_ty.NadaType(
            object=proto_ty.Object(
                fields=[
                    proto_ty.ObjectEntry(name=k, type=v) for k, v in fields.items()
                ],
            )
        )

    def instantiate(self, child_or_value):
        return Object(child_or_value, self.types)


@dataclass
class Object(DslType):
    """The Object type"""

    types: Dict[str, Any]

    def __init__(self, child, types: Dict[str, Any]):
        self.types = types
        self.child = child
        super().__init__(self.child)

    @classmethod
    def new(cls, values: Dict[str, Any]) -> "Object":
        """Constructs a new Object."""
        types = {key: value.type() for key, value in values.items()}
        return Object(
            types=types,
            child=ObjectNew(
                child=values,
                source_ref=SourceRef.back_frame(),
            ),
        )

    def __getattr__(self, attr: str) -> DslType:
        if attr not in self.types:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{attr}'"
            )

        accessor = ObjectAccessor(
            key=attr,
            child=self,
            source_ref=SourceRef.back_frame(),
        )

        return _generate_accessor(self.types[attr], accessor)

    def type(self):
        """Metatype for Object"""
        return ObjectType(types=self.types)


@dataclass
class ObjectAccessor:
    """Accessor for Object"""

    child: Object
    key: str
    source_ref: SourceRef

    def __init__(
        self,
        child: Object,
        key: str,
        source_ref: SourceRef,
    ):
        self.id = OperationId.next()
        self.child = child
        self.key = key
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store this accessor in the AST."""
        AST_OPERATIONS[self.id] = ObjectAccessorASTOperation(
            id=self.id,
            source=self.child.child.id,
            key=self.key,
            source_ref=self.source_ref,
            ty=ty,
        )


class Zip:
    """The Zip operation."""

    def __init__(self, left: AllTypes, right: AllTypes, source_ref: SourceRef):
        self.id = OperationId.next()
        self.left = left
        self.right = right
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store a Zip object in the AST."""
        AST_OPERATIONS[self.id] = BinaryASTOperation(
            id=self.id,
            variant=proto_op.BinaryOperationVariant.ZIP,
            left=self.left.child.id,
            right=self.right.child.id,
            source_ref=self.source_ref,
            ty=ty,
        )


class Unzip:
    """The Unzip operation."""

    def __init__(self, child: AllTypes, source_ref: SourceRef):
        self.id = OperationId.next()
        self.child = child
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store an Unzip object in the AST."""
        AST_OPERATIONS[self.id] = UnaryASTOperation(
            id=self.id,
            variant=proto_op.UnaryOperationVariant.UNZIP,
            child=self.child.child.id,
            source_ref=self.source_ref,
            ty=ty,
        )


class InnerProduct:
    """Inner product of two arrays."""

    def __init__(self, left: AllTypes, right: AllTypes, source_ref: SourceRef):
        self.id = OperationId.next()
        self.left = left
        self.right = right
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store the InnerProduct object in the AST."""
        AST_OPERATIONS[self.id] = BinaryASTOperation(
            id=self.id,
            variant=proto_op.BinaryOperationVariant.INNER_PRODUCT,
            left=self.left.child.id,
            right=self.right.child.id,
            source_ref=self.source_ref,
            ty=ty,
        )


class ArrayType(NadaType):
    """Marker type for arrays."""

    is_compound = True

    def __init__(self, contained_type: NadaType, size: int):
        self.contained_type = contained_type
        self.size = size

    def to_mir(self) -> proto_ty.NadaType:
        """Convert this generic type into a MIR Nada type."""
        return proto_ty.NadaType(
            array=proto_ty.Array(
                size=self.size, contained_type=self.contained_type.to_mir()
            )
        )

    def instantiate(self, child_or_value):
        return Array(child_or_value, self.size, self.contained_type)


@dataclass
class Array(Generic[T], DslType):
    """Nada Array type.

    This is the representation of arrays in Nada MIR.
    Arrays have public, fixed size at compile time.

    Attributes
    ----------
    contained_type: T
        The type of the array
    child:
        The optional child operation
    size: int
        The size of the array
    """

    contained_type: NadaType
    size: int

    def __init__(self, child, size: int, contained_type: NadaType = None):
        self.contained_type = (
            contained_type if contained_type is not None else child.type()
        )

        self.size = size
        self.child = (
            child if contained_type is not None else getattr(child, "child", None)
        )
        if self.child is not None:
            self.child.store_in_ast(self.type().to_mir())

    def __iter__(self):
        raise NotAllowedException(
            "Cannot loop over a Nada Array, use functional style Array operations (map, reduce, zip)."
        )

    def check_not_constant(self, ty):
        """Checks that a type is not a constant"""
        if ty.is_constant:
            raise NotAllowedException(
                "functors (map and reduce) can't be called with constant args"
            )

    def map(self: "Array[T]", function) -> "Array":
        """The map operation for Arrays."""
        self.check_not_constant(self.contained_type)
        nada_function = create_nada_fn(function, args_ty=[self.contained_type])
        return Array(
            size=self.size,
            contained_type=nada_function.return_type,
            child=Map(child=self, fn=nada_function, source_ref=SourceRef.back_frame()),
        )

    def reduce(self: "Array[T]", function, initial: R) -> R:
        """The Reduce operation for arrays."""
        self.check_not_constant(self.contained_type)
        self.check_not_constant(initial.type())
        function = create_nada_fn(
            function, args_ty=[initial.type(), self.contained_type]
        )
        return function.return_type.instantiate(
            Reduce(
                child=self,
                fn=function,
                initial=initial,
                source_ref=SourceRef.back_frame(),
            )
        )

    def zip(self: "Array[T]", other: "Array[U]") -> "Array[Tuple[T, U]]":
        """The Zip operation for Arrays."""
        if self.size != other.size:
            raise IncompatibleTypesError("Cannot zip arrays of different size")
        return Array(
            size=self.size,
            contained_type=TupleType(
                left_type=self.contained_type,
                right_type=other.contained_type,
            ),
            child=Zip(left=self, right=other, source_ref=SourceRef.back_frame()),
        )

    def inner_product(self: "Array[T]", other: "Array[T]") -> T:
        """The child product operation for arrays"""
        if self.size != other.size:
            raise IncompatibleTypesError(
                "Cannot do child product of arrays of different size"
            )

        if is_primitive_integer(self.contained_type) and is_primitive_integer(
            other.contained_type
        ):
            return self.contained_type.instantiate(
                InnerProduct(left=self, right=other, source_ref=SourceRef.back_frame())
            )  # type: ignore

        raise InvalidTypeError(
            "Inner product is only implemented for arrays of integer types"
        )

    @classmethod
    def new(cls, *args) -> "Array[T]":
        """Constructs a new Array."""
        if len(args) == 0:
            raise ValueError("At least one value is required")

        first_arg = args[0]
        if not all(isinstance(arg, type(first_arg)) for arg in args):
            raise TypeError("All arguments must be of the same type")

        return Array(
            contained_type=first_arg.type(),
            size=len(args),
            child=ArrayNew(
                child=args,
                source_ref=SourceRef.back_frame(),
            ),
        )

    def type(self):
        """Metatype for Array"""
        return ArrayType(self.contained_type, self.size)


@dataclass
class TupleNew(Generic[T, U]):
    """MIR Tuple new operation.

    Represents the creation of a new Tuple.
    """

    child: typing.Tuple[T, U]
    source_ref: SourceRef

    def __init__(self, child: typing.Tuple[T, U], source_ref: SourceRef):
        self.id = OperationId.next()
        self.child = child
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store this TupleNew in the AST."""
        AST_OPERATIONS[self.id] = NewASTOperation(
            id=self.id,
            name=self.__class__.__name__,
            elements=[element.child.id for element in self.child],
            source_ref=self.source_ref,
            ty=ty,
        )


@dataclass
class NTupleNew:
    """MIR NTuple new operation.

    Represents the creation of a new Tuple.
    """

    child: List[DslType]
    source_ref: SourceRef

    def __init__(self, child: List[DslType], source_ref: SourceRef):
        self.id = OperationId.next()
        self.child = child
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store this NTupleNew in the AST."""
        AST_OPERATIONS[self.id] = NewASTOperation(
            id=self.id,
            name=self.__class__.__name__,
            elements=[element.child.id for element in self.child],
            source_ref=self.source_ref,
            ty=ty,
        )


@dataclass
class ObjectNew:
    """MIR Object new operation.

    Represents the creation of a new Object.
    """

    child: Dict[str, DslType]
    source_ref: SourceRef

    def __init__(self, child: Dict[str, DslType], source_ref: SourceRef):
        self.id = OperationId.next()
        self.child = child
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store this Object in the AST."""
        AST_OPERATIONS[self.id] = NewASTOperation(
            id=self.id,
            name=self.__class__.__name__,
            elements=[element.child.id for element in self.child.values()],
            source_ref=self.source_ref,
            ty=ty,
        )


def unzip(array: Array[Tuple[T, R]]) -> Tuple[Array[T], Array[R]]:
    """The Unzip operation for Arrays."""
    right_type = ArrayType(
        contained_type=array.contained_type.right_type, size=array.size
    )
    left_type = ArrayType(
        contained_type=array.contained_type.left_type, size=array.size
    )

    return Tuple(
        right_type=right_type,
        left_type=left_type,
        child=Unzip(child=array, source_ref=SourceRef.back_frame()),
    )


@dataclass
class ArrayNew(Generic[T]):
    """MIR Array new operation"""

    child: List[T]
    source_ref: SourceRef

    def __init__(self, child: List[T], source_ref: SourceRef):
        self.id = OperationId.next()
        self.child = child
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store this ArrayNew object in the AST."""
        AST_OPERATIONS[self.id] = NewASTOperation(
            id=self.id,
            name=self.__class__.__name__,
            elements=[element.child.id for element in self.child],
            source_ref=self.source_ref,
            ty=ty,
        )



"""
Class definitions corresponding to Nada operations.
"""

from dataclasses import dataclass
from nada_mir_proto.nillion.nada.types import v1 as proto_ty
from nada_mir_proto.nillion.nada.operations import v1 as proto_op

from nada_dsl import SourceRef
from nada_dsl.ast_util import (
    AST_OPERATIONS,
    BinaryASTOperation,
    IfElseASTOperation,
    RandomASTOperation,
    UnaryASTOperation,
    OperationId,
)
from nada_dsl.nada_types import AllTypes


class BinaryOperation:
    """Superclass of all the binary operations."""

    variant: proto_op.BinaryOperationVariant

    def __init__(self, left: AllTypes, right: AllTypes, source_ref: SourceRef):
        self.id = OperationId.next()
        self.left = left
        self.right = right
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store object in AST"""
        AST_OPERATIONS[self.id] = BinaryASTOperation(
            id=self.id,
            variant=self.variant,
            left=self.left.child.id,
            right=self.right.child.id,
            source_ref=self.source_ref,
            ty=ty,
        )


class UnaryOperation:
    """Superclass of all the unary operations."""

    variant: proto_op.UnaryOperationVariant

    def __init__(self, child: AllTypes, source_ref: SourceRef):
        self.id = OperationId.next()
        self.child = child
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store object in AST."""
        AST_OPERATIONS[self.id] = UnaryASTOperation(
            id=self.id,
            variant=self.variant,
            child=self.child.child.id,
            source_ref=self.source_ref,
            ty=ty,
        )


class Addition(BinaryOperation):
    """Addition operation"""

    variant = proto_op.BinaryOperationVariant.ADDITION


class Subtraction(BinaryOperation):
    """Subtraction operation."""

    variant = proto_op.BinaryOperationVariant.SUBTRACTION


class Multiplication(BinaryOperation):
    """Multiplication operation"""

    variant = proto_op.BinaryOperationVariant.MULTIPLICATION


class Division(BinaryOperation):
    """Division operation"""

    variant = proto_op.BinaryOperationVariant.DIVISION


class Modulo(BinaryOperation):
    """Modulo operation"""

    variant = proto_op.BinaryOperationVariant.MODULO


class Power(BinaryOperation):
    """Power operation"""

    variant = proto_op.BinaryOperationVariant.POWER


class RightShift(BinaryOperation):
    """Right shift (>>) operation."""

    variant = proto_op.BinaryOperationVariant.RIGHT_SHIFT


class LeftShift(BinaryOperation):
    """Left shift (<<)operation."""

    variant = proto_op.BinaryOperationVariant.LEFT_SHIFT


class LessThan(BinaryOperation):
    """Less than (<) operation"""

    variant = proto_op.BinaryOperationVariant.LESS_THAN


class GreaterThan(BinaryOperation):
    """Greater than (>) operation."""

    variant = proto_op.BinaryOperationVariant.GREATER_THAN


class LessOrEqualThan(BinaryOperation):
    """Less or equal (<=) operation."""

    variant = proto_op.BinaryOperationVariant.LESS_EQ


class GreaterOrEqualThan(BinaryOperation):
    """Greater or equal (>=) operation."""

    variant = proto_op.BinaryOperationVariant.GREATER_EQ


class Equals(BinaryOperation):
    """Equals (==) operation"""

    variant = proto_op.BinaryOperationVariant.EQUALS


class NotEquals(BinaryOperation):
    """Not equals (!=) operation."""

    variant = proto_op.BinaryOperationVariant.NOT_EQUALS


class PublicOutputEquality(BinaryOperation):
    """Public output equality operation."""

    variant = proto_op.BinaryOperationVariant.EQUALS_PUBLIC_OUTPUT


class BooleanAnd(BinaryOperation):
    """Boolean AND (&) operation."""

    variant = proto_op.BinaryOperationVariant.BOOL_AND


class BooleanOr(BinaryOperation):
    """Boolean OR (|) operation."""

    variant = proto_op.BinaryOperationVariant.BOOL_OR


class BooleanXor(BinaryOperation):
    """Boolean XOR (^) operation."""

    variant = proto_op.BinaryOperationVariant.BOOL_XOR


class Random:
    """Random operation."""

    source_ref: SourceRef

    def __init__(self, source_ref):
        self.id = OperationId.next()
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store object in AST."""
        AST_OPERATIONS[self.id] = RandomASTOperation(
            id=self.id, ty=ty, source_ref=self.source_ref
        )


@dataclass
class IfElse:
    """
    cond.if_else(left, right)
    """

    this: AllTypes  # cond
    arg_0: AllTypes  # left
    arg_1: AllTypes  # right
    source_ref: SourceRef

    def __init__(
        self, this: AllTypes, arg_0: AllTypes, arg_1: AllTypes, source_ref: SourceRef
    ):
        self.id = OperationId.next()
        self.this = this
        self.arg_0 = arg_0
        self.arg_1 = arg_1
        self.source_ref = source_ref

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store object in AST."""
        AST_OPERATIONS[self.id] = IfElseASTOperation(
            id=self.id,
            condition=self.this.child.id,
            true_branch_child=self.arg_0.child.id,
            false_branch_child=self.arg_1.child.id,
            ty=ty,
            source_ref=self.source_ref,
        )


class Reveal(UnaryOperation):
    """Reveal (i.e. make public) operation."""

    variant = proto_op.UnaryOperationVariant.REVEAL

    def __init__(self, this: AllTypes, source_ref: SourceRef):
        super().__init__(child=this, source_ref=source_ref)


class TruncPr(BinaryOperation):
    """Probabilistic Truncation operation."""

    variant = proto_op.BinaryOperationVariant.TRUNC_PR


class Not(UnaryOperation):
    """Not (!) Operation"""

    variant = proto_op.UnaryOperationVariant.NOT

    def __init__(self, this: AllTypes, source_ref: SourceRef):
        super().__init__(child=this, source_ref=source_ref)


class EcdsaSign(BinaryOperation):
    """Ecdsa signing operation."""

    variant = proto_op.BinaryOperationVariant.ECDSA_SIGN



"""Export classes and functions for Nada DSL auditing component."""

import argparse

from nada_dsl.audit.abstract import *
from nada_dsl.audit.report import html
from nada_dsl.audit.strict import strict


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument("path", nargs=1, help="Nada DSL source file path")
    parser.add_argument("--strict", action="store_true", required=True)
    args = parser.parse_args()
    path = args.path[0]

    with open(path, "r", encoding="UTF-8") as file:
        source = file.read()
        report = strict(source)

    with open(path[:-2] + "html", "w", encoding="UTF-8") as file:
        file.write(html(report))



# pylint:disable=W0401,W0614
"""The Nada Scalar type definitions."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Union, TypeVar
from typing_extensions import Self
from betterproto.lib.google.protobuf import Empty

from nada_mir_proto.nillion.nada.types import v1 as proto_ty

from nada_dsl.operations import *
from nada_dsl.program_io import Literal
from nada_dsl import SourceRef
from . import DslType, Mode, BaseType, OperationType


# Constant dictionary that stores all the Nada types and is use to
# convert from the (mode, base_type) representation to the concrete Nada type
# (Integer, SecretBoolean,...)

# pylint: disable=invalid-name
_AnyScalarType = TypeVar(
    "_AnyScalarType",
    "Integer",
    "UnsignedInteger",
    "Boolean",
    "PublicInteger",
    "PublicUnsignedInteger",
    "PublicBoolean",
    "SecretInteger",
    "SecretUnsignedInteger",
    "SecretBoolean",
)
# pylint: enable=invalid-name

AnyScalarType = Union[
    "Integer",
    "UnsignedInteger",
    "Boolean",
    "PublicInteger",
    "PublicUnsignedInteger",
    "PublicBoolean",
    "SecretInteger",
    "SecretUnsignedInteger",
    "SecretBoolean",
]

# pylint: disable=global-variable-not-assigned
SCALAR_TYPES: dict[tuple[Mode, BaseType], type[AnyScalarType]] = {}

AnyBoolean = Union["Boolean", "PublicBoolean", "SecretBoolean"]


class ScalarDslType(DslType):
    """The Nada Scalar type.
    This is the super class for all scalar types in Nada.
    These are:
        - Boolean, PublicBoolean, SecretBoolean
        - Integer, PublicInteger, SecretInteger
        - UnsignedInteger, PublicUnsignedInteger, SecretUnsignedInteger
    `ScalarType` provides common operation implementations for all the scalar types
    based on the typing rules of the Nada model.
    """

    def __init__(self, child: OperationType, base_type: BaseType, mode: Mode):
        super().__init__(child=child)
        self.base_type = base_type
        self.mode = mode

    def __eq__(self, other) -> AnyBoolean:  # type: ignore
        return equals_operation(
            "Equals", "==", self, other, lambda lhs, rhs: lhs == rhs
        )

    def __ne__(self, other) -> AnyBoolean:  # type: ignore
        return equals_operation(
            "NotEquals", "!=", self, other, lambda lhs, rhs: lhs != rhs
        )

    def to_public(self) -> Self:
        """Convert this scalar type into a public variable."""
        return self

    @classmethod
    def is_scalar(cls) -> bool:
        return True


def equals_operation(
    operation, operator, left: ScalarDslType, right: ScalarDslType, f
) -> AnyBoolean:
    """This function is an abstraction for the equality operations"""
    base_type = left.base_type
    if base_type != right.base_type:
        raise TypeError(f"Invalid operation: {left} {operator} {right}")
    mode = Mode(max([left.mode.value, right.mode.value]))
    match mode:
        case Mode.CONSTANT:
            return Boolean(value=bool(f(left.value, right.value)))
        case Mode.PUBLIC:
            operation = globals()[operation](
                left=left, right=right, source_ref=SourceRef.back_frame().back_frame()
            )
            return PublicBoolean(child=operation)
        case Mode.SECRET:
            operation = globals()[operation](
                left=left, right=right, source_ref=SourceRef.back_frame().back_frame()
            )
            return SecretBoolean(child=operation)


def register_scalar_type(mode: Mode, base_type: BaseType):
    """Decorator used to register a new scalar type in the `SCALAR_TYPES` dictionary."""

    def decorator(scalar_type: type[_AnyScalarType]) -> type[_AnyScalarType]:
        SCALAR_TYPES[(mode, base_type)] = scalar_type
        scalar_type.mode = mode
        scalar_type.base_type = base_type
        return scalar_type

    return decorator


def new_scalar_type(mode: Mode, base_type: BaseType) -> type[AnyScalarType]:
    """Returns the corresponding MIR Nada Type"""
    global SCALAR_TYPES
    return SCALAR_TYPES[(mode, base_type)]


class NumericDslType(ScalarDslType):
    """The superclass of all the numeric types in Nada:
    - Integer, PublicInteger, SecretInteger
    - UnsignedInteger, PublicUnsignedInteger, SecretUnsignedInteger
    It provides common operation implementations for all the numeric types based
    on the typing rules of the Nada model.
    """

    value: int

    def __add__(self, other):
        return binary_arithmetic_operation(
            "Addition", "+", self, other, lambda lhs, rhs: lhs + rhs
        )

    def __sub__(self, other):
        return binary_arithmetic_operation(
            "Subtraction", "-", self, other, lambda lhs, rhs: lhs - rhs
        )

    def __mul__(self, other):
        return binary_arithmetic_operation(
            "Multiplication", "*", self, other, lambda lhs, rhs: lhs * rhs
        )

    def __truediv__(self, other):
        return binary_arithmetic_operation(
            "Division", "/", self, other, lambda lhs, rhs: lhs / rhs
        )

    def __mod__(self, other):
        return binary_arithmetic_operation(
            "Modulo", "%", self, other, lambda lhs, rhs: lhs % rhs
        )

    def __pow__(self, other):
        base_type = self.base_type
        if base_type != other.base_type or not (
            base_type in (BaseType.INTEGER, BaseType.UNSIGNED_INTEGER)
        ):
            raise TypeError(f"Invalid operation: {self} ** {other}")
        mode = Mode(max([self.mode.value, other.mode.value]))
        if mode == Mode.CONSTANT:
            return new_scalar_type(mode, base_type)(self.value**other.value)
        if mode == Mode.PUBLIC:
            child = Power(left=self, right=other, source_ref=SourceRef.back_frame())
            return new_scalar_type(mode, base_type)(child)
        raise TypeError(f"Invalid operation: {self} ** {other}")

    def __lshift__(self, other):
        return shift_operation(
            "LeftShift", "<<", self, other, lambda lhs, rhs: lhs << rhs
        )

    def __rshift__(self, other):
        return shift_operation(
            "RightShift", ">>", self, other, lambda lhs, rhs: lhs >> rhs
        )

    def __lt__(self, other) -> AnyBoolean:
        return binary_relational_operation(
            "LessThan", "<", self, other, lambda lhs, rhs: lhs < rhs
        )

    def __gt__(self, other) -> AnyBoolean:
        return binary_relational_operation(
            "GreaterThan", ">", self, other, lambda lhs, rhs: lhs > rhs
        )

    def __le__(self, other) -> AnyBoolean:
        return binary_relational_operation(
            "LessOrEqualThan", "<=", self, other, lambda lhs, rhs: lhs <= rhs
        )

    def __ge__(self, other) -> AnyBoolean:
        return binary_relational_operation(
            "GreaterOrEqualThan", ">=", self, other, lambda lhs, rhs: lhs >= rhs
        )

    def __radd__(self, other):
        """This adds support for builtin `sum` operation for numeric types."""
        if isinstance(other, int):
            other_type = new_scalar_type(mode=Mode.CONSTANT, base_type=self.base_type)(
                other
            )
            return self.__add__(other_type)

        return self.__add__(other)


def binary_arithmetic_operation(
    operation, operator, left: ScalarDslType, right: ScalarDslType, f
) -> ScalarDslType:
    """This function is an abstraction for the binary arithmetic operations.

    Arithmetic operations apply to Numeric types only in Nada."""
    base_type = left.base_type
    if base_type != right.base_type or not base_type.is_numeric():
        raise TypeError(f"Invalid operation: {left} {operator} {right}")
    mode = Mode(max([left.mode.value, right.mode.value]))
    match mode:
        case Mode.CONSTANT:
            return new_scalar_type(mode, base_type)(f(left.value, right.value))
        case Mode.PUBLIC | Mode.SECRET:
            child = globals()[operation](
                left=left, right=right, source_ref=SourceRef.back_frame().back_frame()
            )
            return new_scalar_type(mode, base_type)(child)


def shift_operation(
    operation, operator, left: ScalarDslType, right: ScalarDslType, f
) -> ScalarDslType:
    """This function is an abstraction for the shift operations"""
    base_type = left.base_type
    right_base_type = right.base_type
    if not base_type.is_numeric() or not right_base_type == BaseType.UNSIGNED_INTEGER:
        raise TypeError(f"Invalid operation: {left} {operator} {right}")
    right_mode = right.mode
    if right_mode == Mode.SECRET:
        raise TypeError(f"Invalid operation: {left} {operator} {right}")
    mode = Mode(max([left.mode.value, right_mode.value]))
    match mode:
        case Mode.CONSTANT:
            return new_scalar_type(mode, base_type)(f(left.value, right.value))
        case Mode.PUBLIC | Mode.SECRET:
            child = globals()[operation](
                left=left, right=right, source_ref=SourceRef.back_frame().back_frame()
            )
            return new_scalar_type(mode, base_type)(child)


def binary_relational_operation(
    operation, operator, left: ScalarDslType, right: ScalarDslType, f
) -> AnyBoolean:
    """This function is an abstraction for the binary relational operations"""
    base_type = left.base_type
    if base_type != right.base_type or not base_type.is_numeric():
        raise TypeError(f"Invalid operation: {left} {operator} {right}")
    mode = Mode(max([left.mode.value, right.mode.value]))
    match mode:
        case Mode.CONSTANT:
            return new_scalar_type(mode, BaseType.BOOLEAN)(f(left.value, right.value))  # type: ignore
        case Mode.PUBLIC | Mode.SECRET:
            child = globals()[operation](
                left=left, right=right, source_ref=SourceRef.back_frame().back_frame()
            )
            return new_scalar_type(mode, BaseType.BOOLEAN)(child)  # type: ignore


def public_equals_operation(
    left: ScalarDslType, right: ScalarDslType
) -> "PublicBoolean":
    """This function is an abstraction for the public_equals operation for all types."""
    base_type = left.base_type
    if base_type != right.base_type:
        raise TypeError(f"Invalid operation: {left}.public_equals({right})")
    if Mode.CONSTANT in (left.mode, right.mode):
        raise TypeError(f"Invalid operation: {left}.public_equals({right})")

    return PublicBoolean(
        child=PublicOutputEquality(
            left=left, right=right, source_ref=SourceRef.back_frame().back_frame()
        )  # type: ignore
    )


class BooleanDslType(ScalarDslType):
    """This abstraction represents all boolean types:
    - Boolean, PublicBoolean, SecretBoolean
    It provides common operation implementations for all the boolean types, defined above.
    """

    def __and__(self, other):
        return binary_logical_operation(
            "BooleanAnd", "&", self, other, lambda lhs, rhs: lhs & rhs
        )

    def __or__(self, other):
        return binary_logical_operation(
            "BooleanOr", "|", self, other, lambda lhs, rhs: lhs | rhs
        )

    def __xor__(self, other):
        return binary_logical_operation(
            "BooleanXor", "^", self, other, lambda lhs, rhs: lhs ^ rhs
        )

    def if_else(self, arg_0: _AnyScalarType, arg_1: _AnyScalarType) -> _AnyScalarType:
        """This function implements the function 'if_else' for every class that extends 'BooleanType'."""
        base_type = arg_0.base_type
        if (
            base_type != arg_1.base_type
            or base_type == BaseType.BOOLEAN
            or self.mode == Mode.CONSTANT
        ):
            raise TypeError(f"Invalid operation: {self}.IfElse({arg_0}, {arg_1})")
        mode = Mode(max([self.mode.value, arg_0.mode.value, arg_1.mode.value]))
        child = IfElse(
            this=self, arg_0=arg_0, arg_1=arg_1, source_ref=SourceRef.back_frame()
        )
        if mode == Mode.CONSTANT:
            mode = Mode.PUBLIC
        return new_scalar_type(mode, base_type)(child)


def binary_logical_operation(
    operation, operator, left: ScalarDslType, right: ScalarDslType, f
) -> ScalarDslType:
    """This function is an abstraction for the logical operations."""
    base_type = left.base_type
    if base_type != right.base_type or not base_type == BaseType.BOOLEAN:
        raise TypeError(f"Invalid operation: {left} {operator} {right}")
    mode = Mode(max([left.mode.value, right.mode.value]))
    if mode == Mode.CONSTANT:
        return Boolean(value=bool(f(left.value, right.value)))
    if mode == Mode.PUBLIC:
        operation = globals()[operation](
            left=left, right=right, source_ref=SourceRef.back_frame().back_frame()
        )
        return PublicBoolean(child=operation)

    operation = globals()[operation](
        left=left, right=right, source_ref=SourceRef.back_frame().back_frame()
    )
    return SecretBoolean(child=operation)


class NadaType(ABC):
    """Abstract meta type"""

    is_constant = False
    is_scalar = False
    is_compound = False

    @abstractmethod
    def instantiate(self, child_or_value):
        """Creates a value corresponding to this meta type"""

    @abstractmethod
    def to_mir(self) -> proto_ty.NadaType:
        """Returns a MIR representation of this meta type"""


class TypePassthroughMixin(NadaType):
    """Mixin for meta types"""

    def instantiate(self, child_or_value):
        """Creates a value corresponding to this meta type"""
        return self.ty(child_or_value)

    def to_mir(self) -> proto_ty.NadaType:
        return proto_ty.NadaType(**{self.proto_ty: Empty()})


@register_scalar_type(Mode.CONSTANT, BaseType.INTEGER)
class Integer(NumericDslType):
    """The Nada Integer type.

    Represents a constant (literal) integer."""

    def __init__(self, value):
        value = int(value)
        super().__init__(
            Literal(value=value, source_ref=SourceRef.back_frame()),
            BaseType.INTEGER,
            Mode.CONSTANT,
        )
        self.value = value

    def __eq__(self, other) -> AnyBoolean:
        return ScalarDslType.__eq__(self, other)

    @classmethod
    def is_literal(cls) -> bool:
        return True

    def type(self):
        return IntegerType()


class IntegerType(TypePassthroughMixin):
    """Meta type for integers"""

    ty = Integer
    is_constant = True
    is_scalar = True
    proto_ty = "integer"


@dataclass
@register_scalar_type(Mode.CONSTANT, BaseType.UNSIGNED_INTEGER)
class UnsignedInteger(NumericDslType):
    """The Nada Unsigned Integer type.

    Represents a constant (literal) unsigned integer."""

    value: int

    def __init__(self, value):
        value = int(value)
        super().__init__(
            Literal(value=value, source_ref=SourceRef.back_frame()),
            BaseType.UNSIGNED_INTEGER,
            Mode.CONSTANT,
        )
        self.value = value

    def __eq__(self, other) -> AnyBoolean:
        return ScalarDslType.__eq__(self, other)

    @classmethod
    def is_literal(cls) -> bool:
        return True

    def type(self):
        return UnsignedIntegerType()


class UnsignedIntegerType(TypePassthroughMixin):
    """Meta type for unsigned integers"""

    ty = UnsignedInteger
    is_constant = True
    is_scalar = True
    proto_ty = "unsigned_integer"


@register_scalar_type(Mode.CONSTANT, BaseType.BOOLEAN)
class Boolean(BooleanDslType):
    """The Nada Boolean type.

    Represents a constant (literal) boolean."""

    value: bool

    def __init__(self, value):
        value = bool(value)
        super().__init__(
            Literal(value=value, source_ref=SourceRef.back_frame()),
            BaseType.BOOLEAN,
            Mode.CONSTANT,
        )
        self.value = value

    def __bool__(self) -> bool:
        return self.value

    def __eq__(self, other) -> AnyBoolean:
        return ScalarDslType.__eq__(self, other)

    def __invert__(self: "Boolean") -> "Boolean":
        return Boolean(value=bool(not self.value))

    @classmethod
    def is_literal(cls) -> bool:
        return True

    def type(self):
        return BooleanType()


class BooleanType(TypePassthroughMixin):
    """Meta type for booleans"""

    ty = Boolean
    is_constant = True
    is_scalar = True
    proto_ty = "boolean"


@register_scalar_type(Mode.PUBLIC, BaseType.INTEGER)
class PublicInteger(NumericDslType):
    """The Nada Public Unsigned Integer type.

    Represents a public unsigned integer in a program. This is a public variable
    evaluated at runtime."""

    def __init__(self, child: DslType):
        super().__init__(child, BaseType.INTEGER, Mode.PUBLIC)

    def __eq__(self, other) -> AnyBoolean:
        return ScalarDslType.__eq__(self, other)

    def public_equals(
        self, other: Union["PublicInteger", "SecretInteger"]
    ) -> "PublicBoolean":
        """Implementation of public equality for Public integer types."""
        return public_equals_operation(self, other)

    def type(self):
        return PublicIntegerType()


class PublicIntegerType(TypePassthroughMixin):
    """Meta type for public integers"""

    ty = PublicInteger
    is_scalar = True
    proto_ty = "integer"


@register_scalar_type(Mode.PUBLIC, BaseType.UNSIGNED_INTEGER)
class PublicUnsignedInteger(NumericDslType):
    """The Nada Public Integer type.

    Represents a public integer in a program. This is a public variable
    evaluated at runtime."""

    def __init__(self, child: DslType):
        super().__init__(child, BaseType.UNSIGNED_INTEGER, Mode.PUBLIC)

    def __eq__(self, other) -> AnyBoolean:
        return ScalarDslType.__eq__(self, other)

    def public_equals(
        self, other: Union["PublicUnsignedInteger", "SecretUnsignedInteger"]
    ) -> "PublicBoolean":
        """Implementation of public equality for Public unsigned integer types."""
        return public_equals_operation(self, other)

    def type(self):
        return PublicUnsignedIntegerType()


class PublicUnsignedIntegerType(TypePassthroughMixin):
    """Meta type for public unsigned integers"""

    ty = PublicUnsignedInteger
    is_scalar = True
    proto_ty = "unsigned_integer"


@dataclass
@register_scalar_type(Mode.PUBLIC, BaseType.BOOLEAN)
class PublicBoolean(BooleanDslType):
    """The Nada Public Boolean type.

    Represents a public boolean in a program. This is a public variable
    evaluated at runtime."""

    def __init__(self, child: DslType):
        super().__init__(child, BaseType.BOOLEAN, Mode.PUBLIC)

    def __eq__(self, other) -> AnyBoolean:
        return ScalarDslType.__eq__(self, other)

    def __invert__(self: "PublicBoolean") -> "PublicBoolean":
        operation = Not(this=self, source_ref=SourceRef.back_frame())
        return PublicBoolean(child=operation)

    def public_equals(
        self, other: Union["PublicUnsignedInteger", "SecretUnsignedInteger"]
    ) -> "PublicBoolean":
        """Implementation of public equality for Public boolean types."""
        return public_equals_operation(self, other)

    def type(self):
        return PublicBooleanType()


class PublicBooleanType(TypePassthroughMixin):
    """Meta type for public booleans"""

    ty = PublicBoolean
    is_scalar = True
    proto_ty = "boolean"


@dataclass
@register_scalar_type(Mode.SECRET, BaseType.INTEGER)
class SecretInteger(NumericDslType):
    """The Nada secret integer type."""

    def __init__(self, child: DslType):
        super().__init__(child, BaseType.INTEGER, Mode.SECRET)

    def __eq__(self, other) -> AnyBoolean:
        return ScalarDslType.__eq__(self, other)

    def public_equals(
        self, other: Union["PublicInteger", "SecretInteger"]
    ) -> "PublicBoolean":
        """Implementation of public equality for secret integer types."""
        return public_equals_operation(self, other)

    def trunc_pr(
        self, other: Union["PublicUnsignedInteger", "UnsignedInteger"]
    ) -> "SecretInteger":
        """Probabilistic truncation for secret integers."""
        if isinstance(other, UnsignedInteger):
            operation = TruncPr(
                left=self, right=other, source_ref=SourceRef.back_frame()
            )
            return SecretInteger(child=operation)
        if isinstance(other, PublicUnsignedInteger):
            operation = TruncPr(
                left=self, right=other, source_ref=SourceRef.back_frame()
            )
            return SecretInteger(child=operation)

        raise TypeError(f"Invalid operation: {self}.trunc_pr({other})")

    @classmethod
    def random(cls) -> "SecretInteger":
        """Random operation for Secret integers."""
        return SecretInteger(child=Random(source_ref=SourceRef.back_frame()))

    def to_public(self: "SecretInteger") -> "PublicInteger":
        """Convert this secret integer into a public variable."""
        operation = Reveal(this=self, source_ref=SourceRef.back_frame())
        return PublicInteger(child=operation)

    def type(self):
        return SecretIntegerType()


class SecretIntegerType(TypePassthroughMixin):
    """Meta type for secret integers"""

    ty = SecretInteger
    is_scalar = True
    proto_ty = "secret_integer"


@dataclass
@register_scalar_type(Mode.SECRET, BaseType.UNSIGNED_INTEGER)
class SecretUnsignedInteger(NumericDslType):
    """The Nada Secret Unsigned integer type."""

    def __init__(self, child: DslType):
        super().__init__(child, BaseType.UNSIGNED_INTEGER, Mode.SECRET)

    def __eq__(self, other) -> AnyBoolean:
        return ScalarDslType.__eq__(self, other)

    def public_equals(
        self, other: Union["PublicUnsignedInteger", "SecretUnsignedInteger"]
    ) -> "PublicBoolean":
        """The public equality operation."""
        return public_equals_operation(self, other)

    def trunc_pr(
        self, other: Union["PublicUnsignedInteger", "UnsignedInteger"]
    ) -> "SecretUnsignedInteger":
        """Probabilistic truncation operation."""
        if isinstance(other, UnsignedInteger):
            operation = TruncPr(
                left=self, right=other, source_ref=SourceRef.back_frame()
            )
            return SecretUnsignedInteger(child=operation)
        if isinstance(other, PublicUnsignedInteger):
            operation = TruncPr(
                left=self, right=other, source_ref=SourceRef.back_frame()
            )
            return SecretUnsignedInteger(child=operation)

        raise TypeError(f"Invalid operation: {self}.trunc_pr({other})")

    @classmethod
    def random(cls) -> "SecretUnsignedInteger":
        """Generate a random secret unsigned integer."""
        return SecretUnsignedInteger(child=Random(source_ref=SourceRef.back_frame()))

    def to_public(
        self: "SecretUnsignedInteger",
    ) -> "PublicUnsignedInteger":
        """Convert this secret into a public variable."""
        operation = Reveal(this=self, source_ref=SourceRef.back_frame())
        return PublicUnsignedInteger(child=operation)

    def type(self):
        return SecretUnsignedIntegerType()


class SecretUnsignedIntegerType(TypePassthroughMixin):
    """Meta type for secret unsigned integers"""

    ty = SecretUnsignedInteger
    is_scalar = True
    proto_ty = "secret_unsigned_integer"


@dataclass
@register_scalar_type(Mode.SECRET, BaseType.BOOLEAN)
class SecretBoolean(BooleanDslType):
    """The SecretBoolean Nada MIR type."""

    def __init__(self, child: DslType):
        super().__init__(child, BaseType.BOOLEAN, Mode.SECRET)

    def __eq__(self, other) -> AnyBoolean:
        return ScalarDslType.__eq__(self, other)

    def __invert__(self: "SecretBoolean") -> "SecretBoolean":
        operation = Not(this=self, source_ref=SourceRef.back_frame())
        return SecretBoolean(child=operation)

    def to_public(self: "SecretBoolean") -> "PublicBoolean":
        """Convert this secret into a public variable."""
        operation = Reveal(this=self, source_ref=SourceRef.back_frame())
        return PublicBoolean(child=operation)

    @classmethod
    def random(cls) -> "SecretBoolean":
        """Generate a random secret boolean."""
        return SecretBoolean(child=Random(source_ref=SourceRef.back_frame()))

    def type(self):
        return SecretBooleanType()


class SecretBooleanType(TypePassthroughMixin):
    """Meta type for secret booleans"""

    ty = SecretBoolean
    is_scalar = True
    proto_ty = "secret_boolean"


@dataclass
class EcdsaSignature(DslType):
    """The EcdsaSignature Nada MIR type."""

    def __init__(self, child: OperationType):
        super().__init__(child=child)

    def type(self):
        return EcdsaSignatureType()


class EcdsaSignatureType(TypePassthroughMixin):
    """Meta type for EcdsaSignatures"""

    ty = EcdsaSignature
    proto_ty = "ecdsa_signature"


@dataclass
class EcdsaDigestMessage(DslType):
    """The EcdsaDigestMessage Nada MIR type."""

    def __init__(self, child: OperationType):
        super().__init__(child=child)

    def type(self):
        return EcdsaDigestMessageType()


class EcdsaDigestMessageType(TypePassthroughMixin):
    """Meta type for EcdsaDigestMessages"""

    ty = EcdsaDigestMessage
    proto_ty = "ecdsa_digest_message"


@dataclass
class EcdsaPrivateKey(DslType):
    """The EcdsaPrivateKey Nada MIR type."""

    def __init__(self, child: OperationType):
        super().__init__(child=child)

    def ecdsa_sign(self, digest: "EcdsaDigestMessage") -> "EcdsaSignature":
        """Random operation for Secret integers."""
        return EcdsaSignature(
            child=EcdsaSign(left=self, right=digest, source_ref=SourceRef.back_frame())
        )

    def type(self):
        return EcdsaPrivateKeyType()


class EcdsaPrivateKeyType(TypePassthroughMixin):
    """Meta type for EcdsaPrivateKeys"""

    ty = EcdsaPrivateKey
    proto_ty = "ecdsa_private_key"



"""
Timer class

This is a timer class used to measure the performance of different stages
in the compilation process.
"""

from dataclasses import dataclass
import functools
import time
from typing import Dict


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


class Clock:
    """NOP implementation of a Clock"""

    def start(self, timer_name: str):
        """Start the timer"""

    def stop(self, timer_name: str):
        """Stop the timer"""

    def report(self) -> Dict[str, float]:
        """Report"""
        return {}


@dataclass
class DefaultClock(Clock):
    """Default implementation for a clock

    The clock keeps a list of timers and implements functions to keep track of
    different timers in a dictionary.

    """

    timers: Dict[str, float]

    def __init__(self):
        self.timers = {}

    def start(self, timer_name: str):
        if timer_name in self.timers:
            raise TimerError(f"timer {timer_name} already running.")
        self.timers[timer_name] = time.perf_counter()

    def stop(self, timer_name: str):
        if self.timers.get(timer_name) is None:
            raise TimerError(
                f"timer {timer_name} is not running, use start() to start it."
            )
        self.timers[timer_name] = time.perf_counter() - self.timers[timer_name]

    def report(self) -> Dict[str, float]:
        return self.timers


@dataclass
class Timer:
    """Timer.

    Basic Timer implementation
    """

    clock: Clock

    def __init__(self):
        self.clock = Clock()

    def enable(self):
        """Enable the timer."""
        self.clock = DefaultClock()

    def is_enabled(self) -> bool:
        """Returns true if the timer is enabled."""
        return isinstance(self.clock, DefaultClock)

    def start(self, timer_name: str):
        """Starts the timer

        Args:
            timer_name (str): The identifier of the timer.
        """
        self.clock.start(timer_name)

    def stop(self, timer_name: str):
        """Stop the timer"""
        self.clock.stop(timer_name)

    def report(self):
        """Returns the report provided by the clock implementation."""
        return self.clock.report()


# Global timer
timer = Timer()


def add_timer(timer_name: str):
    """Add timer decorator.

    Decorator to simplify adding timers.

    Args:
        func (object): The function being decorated
        timer_name (str): The name of the timer

    """

    def decorator_add_timer(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            timer.start(timer_name)
            value = func(*args, **kwargs)
            timer.stop(timer_name)
            return value

        return wrapper

    return decorator_add_timer



"""Program Input Output utilities.

Define the types used for inputs and outputs in Nada programs.

"""

from dataclasses import dataclass
from typing import Any
from nada_mir_proto.nillion.nada.types import v1 as proto_ty

from nada_dsl.ast_util import (
    AST_OPERATIONS,
    InputASTOperation,
    LiteralASTOperation,
    OperationId,
)
from nada_dsl.errors import InvalidTypeError
from nada_dsl.nada_types import AllTypes, Party
from nada_dsl.nada_types import DslType
from nada_dsl.source_ref import SourceRef


class Input(DslType):
    """
    Represents an input to the computation.

    Attributes:
        name (str): The name of the input.
        party (Party): The party providing the input.
        doc (str): Documentation for the input (default "").
    """

    name: str
    party: Party
    doc: str
    source_ref: SourceRef

    def __init__(self, name, party, doc=""):
        self.id = OperationId.next()
        self.name = name
        self.party = party
        self.doc = doc
        self.child = None
        self.source_ref = SourceRef.back_frame()
        super().__init__(self.child)

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store object in AST"""
        AST_OPERATIONS[self.id] = InputASTOperation(
            id=self.id,
            name=self.name,
            ty=ty,
            party=self.party,
            doc=self.doc,
            source_ref=self.source_ref,
        )


@dataclass
class Literal(DslType):
    """
    Represents a literal value.

    Attributes:
        value (Any): The value of the literal.
    """

    value: Any
    source_ref: SourceRef

    def __init__(self, value, source_ref):
        self.id = OperationId.next()
        self.value = value
        self.source_ref = source_ref
        self.child = None
        super().__init__(self.child)

    def store_in_ast(self, ty: proto_ty.NadaType):
        """Store object in AST"""
        AST_OPERATIONS[self.id] = LiteralASTOperation(
            operation_id=self.id,
            name=self.__class__.__name__,
            ty=ty,
            value=self.value,
            source_ref=self.source_ref,
        )


@dataclass
class Output:
    """
    Represents an output from the computation.

    Attributes:
        child (AllTypes): The type of the output.
        party (Party): The party receiving the output.
        name (str): The name of the output.
    """

    child: AllTypes
    party: Party
    name: str
    source_ref: SourceRef

    def __init__(self, child, name, party):
        self.source_ref = SourceRef.back_frame()
        if not issubclass(type(child), DslType):
            raise InvalidTypeError(
                f"{self.source_ref.file}:{self.source_ref.lineno}: Output value "
                f"{child} of type {type(child)} is not "
                f"a Nada type so it isn't a valid output"
            )
        self.child = child
        self.name = name
        self.party = party



"""
Source reference representation data structure.
"""

# pylint: disable=W0603

import os
from dataclasses import dataclass
from typing import Tuple
import inspect
from nada_mir_proto.nillion.nada.mir import v1 as proto_mir

USED_SOURCES = {}
REFS = []

# Global variable that holds the map of source references
index_map = {}
# Global variable that holds the global counter of source reference identifiers.
next_index = 0  # pylint:disable=C0103


@dataclass
class SourceRef:
    """
    Source reference representation, i.e., a specific location in the source code.
    """

    file: str
    lineno: int
    offset: int
    length: int

    @classmethod
    def back_frame(cls) -> "SourceRef":
        """Get the source reference of the calling frame."""
        backend_frame = inspect.currentframe().f_back.f_back
        lineno = backend_frame.f_lineno
        (offset, length) = SourceRef.try_get_line_info(backend_frame, lineno)
        return cls(
            lineno=lineno,
            offset=offset,
            file=os.path.basename(backend_frame.f_code.co_filename),
            length=length,
        )

    @staticmethod
    def try_get_line_info(backend_frame, lineno) -> Tuple[int, int]:
        """Try to get line information from the source code."""
        # We don't include file sources from nada_dsl package.
        # This is to prevent 'nada_fn' wrongly adding nada_dsl source files from this package.
        if "nada_dsl" in backend_frame.f_code.co_filename:
            return 0, 0
        filename = os.path.basename(backend_frame.f_code.co_filename)

        src = None
        try:
            if filename not in USED_SOURCES:
                with open(
                    f"{backend_frame.f_code.co_filename}", encoding="utf-8"
                ) as file:
                    src = file.read()
                USED_SOURCES[filename] = src
            else:
                src = USED_SOURCES[filename]
        except OSError:
            return 0, 0

        lines = src.splitlines()
        if lineno < len(lines):
            offset = 0
            for i in range(lineno - 1):
                offset += len(lines[i]) + 1
            return offset, len(lines[lineno - 1])

        return 0, 0

    def to_index(self) -> int:
        """Index Source Reference objects.
        Adds the current object into a dict as a key
        as well as an entry in an array, and returns an index to it"""
        global next_index
        key = self.to_key()
        value = self.to_mir()
        if key in index_map:
            return index_map[key]

        index_map[key] = next_index
        REFS.append(value)
        next_index += 1
        return index_map[key]

    def to_mir(self):
        """Convert the SourceRef object to MIR"""
        return proto_mir.SourceRef(
            lineno=self.lineno,
            offset=self.offset,
            file=self.file,
            length=self.length,
        )

    def to_key(self):
        """Convert the current object into the key representation used by 'index_map'"""
        return (self.lineno, self.offset, self.file, self.length)

    @staticmethod
    def get_sources():
        """Get all sources."""
        return USED_SOURCES

    @staticmethod
    def get_refs():
        """Get all refs."""
        return REFS
